# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

DEFAULT_FEATURES = {
    'git',
    'make',
}

DEFAULT_FILES = {
    'requirements',
    'requirements-dev',
    'classifiers',
    'version',
}

YAPF_STYLE_CONFIG = 'pep8'
